/***************************************************************************
 *   Copyright (C) 2007  Antony Martin <martin.antony@yahoo.fr>            *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU General Public License           *
 *   as published by the Free Software Foundation; either version 2        *
 *   of the License, or (at your option) any later version.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA          *
 ***************************************************************************/

/* Cree le : 21 decembre 2006
   derniere modification le 20/03/2008 */

#include <SCE/SCEMinimal.h>

#include <SCE/utils/SCEMatrix.h>
#include <SCE/core/SCECMatrix.h>
#include <SCE/interface/SCECamera.h>


/* revise le 20/02/2008 */
void SCE_Camera_Init (SCE_SCamera *cam)
{
    SCE_Matrix4_LoadIdentity (cam->view);
    SCE_Matrix4_LoadIdentity (cam->proj);
    cam->viewport.x = cam->viewport.y = 0;
    cam->viewport.w = cam->viewport.h = 512; /* NOTE: dimensions de l'ecran */
    SCE_Frustum_Init (&cam->frustum);
}

/* ajoute le 20/03/2008 */
SCE_SCamera* SCE_Camera_Create (void)
{
    SCE_SCamera *cam = NULL;
    SCE_btstart ();
    if (!(cam = SCE_malloc (sizeof *cam)))
        Logger_LogSrc ();
    else
        SCE_Camera_Init (cam);
    SCE_btend ();
    return cam;
}
/* ajoute le 20/03/2008 */
void SCE_Camera_Delete (SCE_SCamera *cam)
{
    SCE_free (cam);
}

void SCE_Camera_SetViewport (SCE_SCamera *cam, int x, int y, int w, int h)
{
    cam->viewport.x = x;
    cam->viewport.y = y;
    cam->viewport.w = w;
    cam->viewport.h = h;
}

/* ajoute le 20/02/2008 */
float* SCE_Camera_GetView (SCE_SCamera *cam)
{
    return cam->view;
}
/* ajoute le 20/02/2008 */
float* SCE_Camera_GetProj (SCE_SCamera *cam)
{
    return cam->proj;
}

/* ajoute le 10/03/2008 */
SCE_SFrustum* SCE_Camera_GetFrustum (SCE_SCamera *cam)
{
    return &cam->frustum;
}

/* ajoute le 10/03/2008 */
void SCE_Camera_UpdateFrustum (SCE_SCamera *cam)
{
    SCE_Frustum_MakeFromMatrices (&cam->frustum, cam->view, cam->proj);
}

/* revise le 20/02/2008 */
void SCE_Camera_Use (SCE_SCamera *cam)
{
    SCE_CViewport (cam->viewport.x, cam->viewport.y,
                   cam->viewport.w, cam->viewport.h);
    SCE_CSetActiveMatrix (SCE_MAT_PROJECTION);
    SCE_CLoadMatrix (cam->proj);  /* Load ou Mult ? */
    SCE_CSetActiveMatrix (SCE_MAT_MODELVIEW);
    SCE_CLoadMatrix (cam->view);  /* Load ou Mult ? */
}
