/***************************************************************************
 *   Copyright (C) 2007  Antony Martin <martin.antony@yahoo.fr>            *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU General Public License           *
 *   as published by the Free Software Foundation; either version 2        *
 *   of the License, or (at your option) any later version.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA          *
 ***************************************************************************/

/* Cree le : 10 juillet 2007
   derniere modification : 07/01/2008 */

#include <SCE/SCEMinimal.h>

#include <SCE/core/SCECMatrix.h>

#include <SCE/interface/SCENode.h>

/* hack pour la suppression d'un noeud et de sa liste... */
static int lists_hack = 1, was = 0;

/* ajoute le 06/01/2008 */
static void SCE_Node_Init (SCE_SNode *node)
{
    node->parent = NULL;
    node->child = NULL;
    SCE_Matrix4_LoadIdentity (node->matrix);
    SCE_Matrix4_LoadIdentity (node->final_matrix);
    node->updated = SCE_FALSE;
    node->data = NULL;
    node->f = NULL;
    node->fparam = NULL;
}

/* ajoute le 10/07/2007 */
/* revise le 04/03/2008 */
SCE_SNode* SCE_Node_Create (void)
{
    SCE_SNode *node = NULL;

    node = SCE_malloc (sizeof *node);
    if (!node)
        Logger_LogSrc ();
    else
    {
        SCE_Node_Init (node);
        node->child = SCE_List_Create (SCE_Node_Delete);
        if (!node->child)
        {
            SCE_free (node), node = NULL;
            Logger_LogSrc ();
        }
    }
    return node;
}

/* ajoute le 10/07/2007 */
/* revise le 06/01/2008 */
void SCE_Node_Delete (void *n)
{
    if (n)
    {
        SCE_SNode *node = n;
        if (lists_hack) /* on le detache de son papa */
            SCE_Node_Detach (node);
        /* TODO: va falloir faire quelque chose avec ses fils...
                 on les ajoute a son ancien papa ? */
        if (!lists_hack)
            was = 1;
        else
            lists_hack = 0;
        SCE_List_Delete (node->child);
        if (!was)
            lists_hack = 1;
        SCE_free (node);
    }
}


/* ajoute le 10/07/2007 */
/* revise le 06/01/2008 */
static void SCE_Node_IsNotUpdated (SCE_SNode *node)
{
    SCE_SListIterator *i = NULL;
    node->updated = SCE_FALSE;
    SCE_List_ForEach (i, node->child)
        SCE_Node_IsNotUpdated (SCE_List_GetData (i));
}


/* ajoute le 07/01/2008 */
int SCE_Node_Attach (SCE_SNode *node, SCE_SNode *child)
{
    SCE_Node_Detach (child); /* impossible d'avoir deux parents */
    if (SCE_List_AddFront (node->child, child) < 0)
    {
        Logger_LogSrc ();
        return SCE_ERROR;
    }
    child->parent = node;
    /* child et ses fils ne sont plus a jour */
    SCE_Node_IsNotUpdated (child);
    return SCE_OK;
}

/* ajoute le 07/01/2008 */
SCE_SNode* SCE_Node_AttachNew (SCE_SNode *node)
{
    SCE_SNode *new = NULL;
    new = SCE_Node_Create ();
    if (!new)
    {
        Logger_LogSrc ();
        return NULL;
    }
    if (SCE_Node_Attach (node, new) < 0)
    {
        SCE_Node_Delete (new);
        Logger_LogSrc ();
        return NULL;
    }
    return new;
}

/* ajoute le 07/01/2008 */
void SCE_Node_Detach (SCE_SNode *node)
{
    if (node->parent)
    {
        SCE_SListIterator *i = NULL;
        i = SCE_List_LocateIterator (node->parent->child, node, NULL);
        SCE_List_Remove (node->parent->child, i);
        node->parent = NULL;
        SCE_Node_IsNotUpdated (node);
    }
}

/* ajoute le 07/01/2008 */
void* SCE_Node_Erase (SCE_SNode *node)
{
    void *data = NULL;
    SCE_Node_Detach (node);
    data = node->data;
    SCE_Node_Delete (node);
    return data;
}

/* ajoute le 12/07/2007 */
/* revise le 07/01/2008 */
int SCE_Node_Insert (SCE_SNode *node, SCE_SNode *child)
{
    if (child->parent)
    {
        if (SCE_Node_Attach (child->parent, node) < 0)
        {
            Logger_LogSrc ();
            return SCE_ERROR;
        }
    }
    if (SCE_Node_Attach (node, child) < 0)
    {
        Logger_LogSrc ();
        return SCE_ERROR;
    }
    return SCE_OK;
}


/* ajoute le 11/07/2007 */
/* revise le 09/03/2008 */
void SCE_Node_MultMatrix (SCE_SNode *node)
{
    SCE_CMultMatrix (node->final_matrix);
}
/* ajoute le 09/01/2008 */
/* revise le 09/03/2008 */
void SCE_Node_LoadMatrix (SCE_SNode *node)
{
    SCE_CLoadMatrix (node->final_matrix);
}

/* ajoute le 10/07/2007 */
/* revise le 07/01/2008 */
float* SCE_Node_GetMatrix (SCE_SNode *node)
{
    SCE_Node_IsNotUpdated (node);
    return node->matrix;
}
/* ajoute le 10/07/2007 */
void SCE_Node_CopyMatrix (SCE_SNode *node, SCE_TMatrix4 m)
{
    SCE_Matrix4_Copy (m, node->matrix);
}

/* ajoute le 11/07/2007 */
/* revise le 07/01/2008 */
float* SCE_Node_GetFinalMatrix (SCE_SNode *node)
{
    SCE_Node_Update (node);
    SCE_Node_IsNotUpdated (node); /* ...et en plus elle n'est plus a jour :D */
    return node->final_matrix;
}
/* ajoute le 11/07/2007 */
/* revise le 23/03/2008 */
void SCE_Node_CopyFinalMatrix (SCE_SNode *node, SCE_TMatrix4 m)
{
    SCE_Node_Update (node);
    SCE_Matrix4_Copy (m, node->final_matrix);
}

/* ajoute le 09/01/2008 */
void SCE_Node_SetCallback (SCE_SNode *node, SCE_FNodeCallback f, void *p)
{
    node->f = f;
    node->fparam = p;
}

/* ajoute le 11/07/2007 */
/* revise le 07/01/2008 */
void SCE_Node_Update (SCE_SNode *node)
{
    if (node->f)
        node->f (node, node->fparam);

    if (!node->updated)
    {
        if (node->parent != NULL)
        {
            /* on update tous les parents */
            SCE_Node_Update (node->parent);

            /* on multiplie la matrice mere par la matrice fille pour
               obtenir la matrice resultante finale */
            SCE_Matrix4_Mul (node->parent->final_matrix, node->matrix,
                             node->final_matrix);
        }
        else
            SCE_Matrix4_Copy (node->final_matrix, node->matrix);

        node->updated = SCE_TRUE;
    }
}

/* ajoute le 07/01/2008 */
void SCE_Node_RecursiveUpdate (SCE_SNode *node)
{
    SCE_SListIterator *i = NULL;
    SCE_Node_Update (node);
    SCE_List_ForEach (i, node->child)
        SCE_Node_RecursiveUpdate (SCE_List_GetData (i));
}


/* ajoute le 12/07/2007 */
int SCE_Node_HaveParent (SCE_SNode *node)
{
    return (node->parent ? SCE_TRUE : SCE_FALSE);
}

/* ajoute le 12/07/2007 */
SCE_SNode* SCE_Node_GetParent (SCE_SNode *node)
{
    return node->parent;
}


/* ajoute le 12/07/2007 */
void* SCE_Node_SetData (SCE_SNode *node, void *data)
{
    void *old_data = node->data;
    node->data = data;
    /* on renvoie les anciennes donnees */
    return old_data;
}

/* ajoute le 12/07/2007 */
void* SCE_Node_GetData (SCE_SNode *node)
{
    return node->data;
}


/* ajoute le 09/01/2008 */
void SCE_Node_Use (SCE_SNode *node)
{
    /*SCE_CSetActiveMatrix (GL_MODELVIEW);*/
    if (node)
    {
        SCE_Node_Update (node);
        SCE_Node_MultMatrix (node);
    }
    else
        SCE_CLoadIdentityMatrix ();
}
