/***************************************************************************
 *   Copyright (C) 2007  Antony Martin <martin.antony@yahoo.fr>            *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU General Public License           *
 *   as published by the Free Software Foundation; either version 2        *
 *   of the License, or (at your option) any later version.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA          *
 ***************************************************************************/

/* Cree le : 21/09/2007
   derniere modification : 11/01/2008 */

#include <SCE/utils/SCEList.h>


/* ajoute le 21/09/2007 */
/* initialise un iterateur */
static void SCE_List_InitIt (SCE_SListIterator *it)
{
    it->next = it->prev = NULL;
    it->data = NULL;
}

/* ajoute le 23/09/2007 */
static SCE_SListIterator* SCE_List_CreateIt (void)
{
    SCE_SListIterator *it = SCE_malloc (sizeof *it);
    if (!it)
        Logger_LogSrc ();
    else
        SCE_List_InitIt (it);
    return it;
}

/* ajoute le 23/09/2007 */
/* revise le 11/01/2008 */
static void SCE_List_DeleteIt (SCE_SListIterator *it)
{
    SCE_free (it);
}


/* ajoute le 21/09/2007 */
/* revise le 02/10/2007 */
void SCE_List_Init (SCE_SList *l)
{
    l->first = NULL;
    l->f = NULL;
}


/* ajoute le 21/09/2007 */
/* revise le 22/09/2007 */
SCE_SList* SCE_List_Create (SCE_FListFreeFunc f)
{
    SCE_SList *l = NULL;

    l = SCE_malloc (sizeof *l);
    if (l == NULL)
    {
        Logger_LogSrc ();
        return NULL;
    }

    SCE_List_Init (l);
    l->f = f;

    return l;
}

/* ajoute le 24/09/2007 */
void SCE_List_Clear (SCE_SList *l)
{
    SCE_SListIterator *it = l->first;
    SCE_SListIterator *tmp = NULL;

    while (it)
    {
        tmp = it;
        it = it->next;
        if (l->f)
            l->f (tmp->data);
        SCE_List_DeleteIt (tmp);
    }

    l->first = NULL;
}

/* ajoute le 21/09/2007 */
/* revise le 24/09/2007 */
void SCE_List_Delete (SCE_SList *l)
{
    if (l)
    {
        SCE_List_Clear (l);
        SCE_free (l);
    }
}


/* ajoute le 22/09/2007 */
/* revise le 08/10/2007 */
int SCE_List_AddFront (SCE_SList *l, void *d)
{
    SCE_SListIterator *it = SCE_List_CreateIt ();
    if (!it)
    {
        Logger_LogSrc ();
        return SCE_ERROR;
    }
    it->data = d;
    it->next = l->first;
    it->prev = NULL;
    if (it->next)
        it->next->prev = it;
    l->first = it;

    return SCE_OK;
}

/* ajoute le 22/09/2007 */
/* revise le 08/10/2007 */
int SCE_List_AddBack (SCE_SList *l, void *d)
{
    SCE_SListIterator *it = SCE_List_CreateIt ();
    if (!it)
    {
        Logger_LogSrc ();
        return SCE_ERROR;
    }
    it->data = d;
    it->next = NULL;
    it->prev = SCE_List_GetLast (l);
    if (it->prev)
        it->prev->next = it;
    else
        l->first = it;

    return SCE_OK;
}

#if 0
/* ajoute le 22/09/2007 */
int SCE_List_Add (SCE_SList*, SCE_SListIterator*, void*)
{
    
}
#endif


/* ajoute le 22/09/2007 */
/* revise le 11/01/2008 */
void* SCE_List_RemoveFront (SCE_SList *l)
{
    void *d = NULL;
    SCE_SListIterator *it = l->first;

    if (it)
    {
        if (it->next)
            it->next->prev = NULL;

        l->first = it->next;
        d = it->data;
        SCE_List_DeleteIt (it);
    }

    return d;
}


/* ajoute le 22/09/2007 */
/* revise le 08/10/2007 */
void* SCE_List_RemoveBack (SCE_SList *l)
{
    void *d = NULL;
    SCE_SListIterator *it = SCE_List_GetLast (l);

    if (it)
    {
        if (it->prev)
            it->prev->next = NULL;
        else
            l->first = NULL;
        d = it->data;
        SCE_List_DeleteIt (it);
    }

    return d;
}

/* ajoute le 22/09/2007 */
/* revise le 01/10/2007 */
void* SCE_List_Remove (SCE_SList *l, SCE_SListIterator *it)
{
    void *d = it->data;

    if (it->prev)
        it->prev->next = it->next;
    else
        l->first = it->next;

    if (it->next)
        it->next->prev = it->prev;

    SCE_List_DeleteIt (it);

    return d;
}



/* ajoute le 22/09/2007 */
/* revise le 25/09/2007 */
void SCE_List_EraseFront (SCE_SList *l)
{
    void *d = SCE_List_RemoveFront (l);
    if (l->f)
        l->f (d);
}

/* ajoute le 22/09/2007 */
/* revise le 25/09/2007 */
void SCE_List_EraseBack (SCE_SList *l)
{
    void *d = SCE_List_RemoveBack (l);
    if (l->f)
        l->f (d);
}

/* ajoute le 22/09/2007 */
/* revise le 01/10/2007 */
void SCE_List_Erase (SCE_SList *l, SCE_SListIterator *it)
{
    void *d = SCE_List_Remove (l, it);
    if (l->f)
        l->f (d);
}


unsigned int SCE_List_GetSize (SCE_SList *l)
{
    unsigned int n = 0;
    SCE_SListIterator *it;
    SCE_List_ForEach (it, l)
        n++;
    return n;
}

/* ajoute le 22/09/2007 */
/* revise le 23/09/2007 */
SCE_SListIterator* SCE_List_GetLast (SCE_SList *l)
{
    SCE_SListIterator *it = l->first;
    if (it)
    {
        while (it->next)
            it = it->next;
    }
    return it;
}

/* ajoute le 24/09/2007 */
void* SCE_List_GetData (SCE_SListIterator *it)
{
    return it->data;
}

/* ajoute le 26/09/2007 */
/* revise le 08/10/2007 */
unsigned int SCE_List_GetIndex (SCE_SListIterator *it)
{
    unsigned int n = 0;
    SCE_List_ForEachPrev (it)
        n++;
    return n - 1;
}

/* ajoute le 26/09/2007 */
SCE_SListIterator* SCE_List_GetIterator (SCE_SList *l, unsigned int n)
{
    SCE_SListIterator *it = NULL;
    unsigned int i = 0;
    
    SCE_List_ForEach (it, l)
    {
        if (i == n)
            return it;
        i++;
    }
    /* iterateur numero 'n' non trouve dans la liste */
    Logger_Log (/*SCE_STACK_OVERFLOW*/SCE_INVALID_ARG);
    return NULL;
}

/* ajoute le 15/10/2007 */
SCE_SListIterator* SCE_List_LocateIterator (SCE_SList *l, void *data,
                                            SCE_FListCompareData f)
{
    SCE_SListIterator *i = NULL;
    SCE_List_ForEach (i, l)
    {
        if ((f && f (data, i->data)) || data == i->data)
            return i;
    }
    return NULL;
}

/* ajoute le 20/10/2007 */
unsigned int SCE_List_LocateIndex (SCE_SList *l, void *data,
                                   SCE_FListCompareData f)
{
    unsigned int j = 0;
    SCE_SListIterator *i = NULL;
    SCE_List_ForEach (i, l)
    {
        if ((f && f (data, i->data)) || data == i->data)
            return j;
        j++;
    }
    return 0;
}

/* ajoute le 25/09/2007 */
void SCE_List_SetFreeFunc (SCE_SList *l, SCE_FListFreeFunc f)
{
    l->f = f;
}
